"""
Functions to prepare input for `pywapor.et_look`, more specifically to
group various parameters in time to create composites. 
"""
import os
import numpy as np
import xarray as xr
import pandas as pd
from itertools import chain
from pywapor.general.logger import log
import numpy as np
from pywapor.general.processing_functions import save_ds, open_ds, remove_ds, adjust_timelim_dtype, log_example_ds
import os
from pywapor.general.reproject import align_pixels
from pywapor.enhancers.apply_enhancers import apply_enhancer
import xarray as xr
import pandas as pd
from datetime import datetime as dt
import dask
import types
import functools
import pywapor.general.levels as levels

dask.config.set(**{'array.slicing.split_large_chunks': True})

def add_times(ds, bins, composite_type):
    """Add times to the time coordinates, so that every bin has at least one
    datapoint.

    Parameters
    ----------
    ds : xr.Dataset
        Datasat for which to check empty bins.
    bins : list
        List of np.datetime64's which are the boundaries of the groups into
        which the variables will grouped.
    composite_type : {"min" | "max" | "mean"}
        Type of composites that will be created based on the data inside each bin.

    Returns
    -------
    xr.Dataset
        Dataset to which time coordinates have been added to assure no empty bins
        exist.
    """
    
    try:
        bin_count = ds.time.groupby_bins("time", bins).count()
        empty_bins = bin_count.sel({"time_bins": bin_count.isnull()}).time_bins
    except ValueError as e:
        if "None of the data falls within bins" in str(e):
            x = [pd.Interval(pd.Timestamp(x0), 
                            pd.Timestamp(x1)) for x0, x1 in zip(bins[:-1], bins[1:])]
            empty_bins = xr.DataArray(x, {"time_bins": x})
        else:
            raise e

    if composite_type == "mean":
        new_t = [x.mid for x in empty_bins.values]
    else:
        new_t1 = [x.left + pd.Timedelta(seconds=1) for x in empty_bins.values]
        new_t2 = [x.right - pd.Timedelta(seconds=1) for x in empty_bins.values]
        new_t = new_t1 + new_t2

    if len(new_t) > 0:
        ds = xr.merge([ds, xr.Dataset({"time": new_t})]).sortby("time")

    return ds

def time_bins(timelim, bin_length):
    """Based on the time limits and the bin length, create the bin boundaries.

    Parameters
    ----------
    timelim : list
        Period for which to prepare data.
    bin_length : int | "DEKAD"
        Length of the bins in days or "DEKAD" for dekadal bins.

    Returns
    -------
    list
        List of np.datetime64's which are the boundaries of the groups into
        which the variables will grouped.
    """
    timelim = adjust_timelim_dtype(timelim)
    sdate = timelim[0]
    edate = timelim[1]
    if bin_length == "DEKAD":
        dekad1 = pd.date_range(sdate - pd.to_timedelta(35, unit='d'), edate + pd.to_timedelta(35, unit='d'), freq = "MS")
        dekad2 = dekad1 + pd.Timedelta(10, unit='D')
        dekad3 = dekad1 + pd.Timedelta(20, unit='D')
        dates = np.sort(np.array([dekad1, dekad2, dekad3]).flatten())
        big_interval = pd.Interval(pd.Timestamp(sdate), pd.Timestamp(edate))
        intervals = [pd.Interval(pd.Timestamp(x), pd.Timestamp(y)) for x,y in zip(dates[:-1], dates[1:])]
        out = np.unique([[np.datetime64(x.right, "ns"), np.datetime64(x.left, "ns")] for x in intervals if x.overlaps(big_interval)]).flatten()
    else:
        days = (edate - sdate).days
        no_periods = int(days // bin_length + 1)
        dates = pd.date_range(sdate, periods = no_periods + 1 , freq = f"{bin_length}D")
        out = dates.to_numpy()
    return out

def main(dss, sources, folder, general_enhancers, bins):
    """Create composites for variables contained in the 'xr.Dataset's in 'dss'.

    Parameters
    ----------
    dss : dict
        Keys are tuples of ('source', 'product_name'), values are xr.Dataset's 
        which will be aligned along the time dimensions.
    sources : dict
        Configuration for each variable and source.
    folder : str
        Path to folder in which to store (intermediate) data.
    general_enhancers : list
        Functions to apply to the xr.Dataset before creating the final
        output, by default "default".
    bins : list
        List of 'np.datetime64's which are the boundaries of the groups into
        which the variables will grouped.

    Returns
    -------
    xr.Dataset
        Dataset with variables grouped into composites.
    """
    # Open unopened netcdf files.
    dss = {**{k: open_ds(v) for k, v in dss.items() if isinstance(v, str)}, 
            **{k:v for k,v in dss.items() if not isinstance(v, str)}}

    final_path = os.path.join(folder, "et_look_in.nc")

    dss2 = dict()

    compositers = {
        "mean": xr.DataArray.mean,
        "min": xr.DataArray.min,
        "max": xr.DataArray.max,
    }

    cleanup = list()

    log.info(f"--> Compositing {len(sources)} variables.").add()

    for i, (var, config) in enumerate(sources.items()):

        spatial_interp = config["spatial_interp"]
        temporal_interp = config["temporal_interp"]
        composite_type = config["composite_type"]

        log.info(f"--> ({i+1}/{len(sources)}) Compositing `{var}` ({composite_type}).").add()
        
        # Align pixels of different products for a single variable together.
        dss_part = [ds[[var]] for ds in dss.values() if var in ds.data_vars]
        dss1, temp_files1 = align_pixels(dss_part, folder, spatial_interp, fn_append = "_step1")
        cleanup.append(temp_files1)

        # Combine different source_products (along time dimension).
        if np.all(["time" in x[var].dims for x in dss1]):
            ds = xr.combine_nested(dss1, concat_dim = "time").chunk({"time": -1}).sortby("time").squeeze()
        elif np.all(["time" not in x[var].dims for x in dss1]):
            ds = xr.concat(dss1, "stacked").median("stacked")
            if len(dss1) > 1:
                log.warning(f"--> Multiple ({len(dss1)}) sources for time-invariant `{var}` found, reducing those with 'median'.")
        else:
            ds = xr.combine_nested([x for x in dss1 if "time" in x[var].dims], concat_dim = "time").chunk({"time": -1}).sortby("time").squeeze()
            log.warning(f"--> Both time-dependent and time-invariant data found for `{var}`, dropping time-invariant data.")

        if "time" in ds.dims:

            if temporal_interp:

                # When combining different products, it is possible to have images 
                # on the exact same date & time. In that case, the median of those images
                # is used. So gaps in one image are filled in with data from the other
                # image(s).
                if np.unique(ds.time).size != ds.time.size:
                    groups = ds.groupby(ds["time"])
                    ds = groups.median(dim = "time")
                    ds = ds.chunk({"time": -1, "y": "auto", "x": "auto"})
                    log.warning(f"--> Multiple `{var}` images for an identical datetime found, reducing those with 'median'.")

                ds = add_times(ds, bins, composite_type = composite_type)
                ds = ds.interpolate_na(dim="time", method = temporal_interp)

            # Make composites.
            ds[var] = compositers[composite_type](ds[var].groupby_bins("time", bins, labels = bins[:-1]))

            # Drop time coordinates.
            ds = ds.drop_vars(["time"])

        # Save output
        dst_path = os.path.join(folder, f"{var}_bin.nc")
        ds = save_ds(ds, dst_path, label = f"Saving `{var}` composites.")
        dss2[var] = ds
        cleanup.append([ds])
        log.sub()

    log.sub()

    # Apply variable specific enhancers
    variables = levels.find_setting(sources, "variable_enhancers")
    for var in variables:
        if var in dss2.keys():
            for func in sources[var]["variable_enhancers"]:
                dss2[var] = func(dss2, var, folder)

    # Align all the variables together.
    example_source = levels.find_setting(sources, "is_example", max_length = 1, min_length = 1)
    if len(example_source) == 1:
        example_ds = dss[example_source[0]]
        log_example_ds(example_ds)
    else:
        log.warning(f"--> No valid example dataset set.")
        example_ds = None
    spatial_interps = [sources[list(x.data_vars)[0]]["spatial_interp"] for x in dss2.values()]
    dss3, temp_files3 = align_pixels(dss2.values(), folder, spatial_interps, example_ds, stack_dim = "time_bins", fn_append = "_step2")
    cleanup.append(temp_files3)
    
    # Merge everything
    ds = xr.merge(dss3)

    # Apply general enhancers.
    for func in general_enhancers:
        ds, label = apply_enhancer(ds, None, func)
        log.info(label)

    while os.path.isfile(final_path):
        final_path = final_path.replace(".nc", "_.nc")

    ds = save_ds(ds, final_path, encoding = "initiate", 
                    label = f"Creating merged file `{os.path.split(final_path)[-1]}`.")
    
    for nc in list(chain.from_iterable(cleanup)):
        remove_ds(nc)

    return ds

if __name__ == "__main__":

    cleanup = False

#     import datetime

#     sources = {
#         "ndvi":         [("MODIS", "MOD13Q1.061"), ("MODIS", "MYD13Q1.061")],
#         "r0":           [("MODIS", "MCD43A3.061")],
#         "lst":          [("MODIS", "MOD11A1.061"), ("MODIS", "MYD11A1.061")],
#         "z":            [("SRTM", "30M")],
#         "p":            [("CHIRPS", "P05")],
#         "ra":           [("MERRA2", "M2T1NXRAD.5.12.4")],
#         "t_air":        [("MERRA2", "M2I1NXASM.5.12.4")],
#         # "t_air_max":    [("MERRA2", "M2I1NXASM.5.12.4")],
#     }

#     diagnostics = { # label          # lat      # lon
#                     "water":	    (29.44977,	30.58215),
#                     "desert":	    (29.12343,	30.51222),
#                     "agriculture":	(29.32301,	30.77599),
#                     "urban":	    (29.30962,	30.84109),
#                     }

#     folder = r"/Users/hmcoerver/Downloads/pywapor_test"
#     latlim = [28.9, 29.7]
#     lonlim = [30.2, 31.2]
#     timelim = [datetime.date(2020, 6, 25), datetime.date(2020, 7, 30)]
#     example_source = ("MODIS", "MOD13Q1.061")
#     bin_length = 4

#     bins = time_bins(timelim, bin_length)
#     dss = collect_sources(folder, sources, latlim, lonlim, [bins[0], bins[-1]])

#     ds = main(dss, sources, example_source, bins, folder, 
#                 diagnostics = None)

#     if diagnostics:
#         _ = main(dss, sources, example_source, bins, folder, diagnostics = diagnostics)


"""Functions to make adjustments to temperature maps.
"""
import numpy as np
from scipy.signal import convolve2d, oaconvolve, fftconvolve
import numpy as np
import xarray as xr
from pywapor.general.logger import log

def template(ds, var, out_var = None):
    """Example enhancer function.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset.
    var : str
        Variable to adjust.
    out_var : str, optional
        variable to store new variable, by default None.

    Returns
    -------
    xr.Dataset
        Output dataset.
    """

    new_data = ds[var]

    if not isinstance(out_var, type(None)):
        ds[out_var] = new_data
    else: 
        ds[var] = new_data

    return ds

def kelvin_to_celsius(ds, var, in_var = None, out_var = None):
    """Convert units of a variable from Kelvin to degrees Celsius.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing `var`.
    var : str
        Variable name.
    in_var : str, optional
        Overwrites `var`, by default None.
    out_var : str, optional
        Instead of overwriting `var`, store the new data in `out_var`, 
        by default None.

    Returns
    -------
    xarray.Dataset
        Adjusted data.
    """
    if not isinstance(in_var, type(None)):
        var = in_var

    if "units" in ds[var].attrs.keys():
        current_unit = ds[var].attrs["units"]
    else:
        current_unit = "unknown"

    if (current_unit == "K") or (current_unit == "unknown"):
        new_data = ds[var] - 273.15
    else:
        new_data = ds[var]

    if "sources" in ds[var].attrs.keys():
        new_data.attrs = {"units": "C",
                            "sources": ds[var].attrs["sources"]}
    else:
        new_data.attrs = {"units": "C"}

    if not isinstance(out_var, type(None)):
        ds[out_var] = new_data
    else: 
        ds[var] = new_data

    return ds

def celsius_to_kelvin(ds, var, out_var = None):
    """Convert units of a variable from Kelvin to degrees Celsius.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing `var`.
    var : str
        Variable name.
    out_var : str, optional
        Instead of overwriting `var`, store the new data in `out_var`, 
        by default None.

    Returns
    -------
    xarray.Dataset
        Adjusted data.
    """

    new_data = ds[var] + 273.15
    if "sources" in ds[var].attrs.keys():
        new_data.attrs = {"units": "K",
                            "sources": ds[var].attrs["sources"]}
    else:
        new_data.attrs = {"unit": "K"}

    if not isinstance(out_var, type(None)):
        ds[out_var] = new_data
    else: 
        ds[var] = new_data

    return ds

def lapse_rate(ds, var, out_var = None, lapse_var = "z", 
                radius = 0.25, lapse_rate = -0.006):
    """Apply a lapse-rate to `var` inside `ds` based on the difference between
    the actual value of `lapse_var` and its local (defined by `radius`) mean.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset with the data.
    var : str
        Name of the variable inside `ds` to apply this function to (y).
    out_var : str, optional
        Instead of overwriting `var`, store the new data in `out_var`, 
        by default None.
    lapse_var : str, optional
        The variable used to calculate dx.
    radius : float, optional
        The local altitude is compared to the local mean altitude, radius
        (in degrees) defines over which area to calculate the mean, by default 0.25.
    lapse_rate : float, optional
        dy/dx, i.e. how much should `var` be adjusted per difference in 
        `lapse_var`, by default -0.006.

    Returns
    -------
    xarray.Dataset
        Adjusted data.
    """
    
    if "t_diff" not in list(ds.keys()):

        log.info(f"--> Calculating local means (r = {radius:.2f}°) of `{lapse_var}`.")
        log.add()

        pixel_size = (np.median(np.diff(ds.x)), 
                        np.median(np.diff(ds.y)))
        dem = ds[lapse_var].values

        missing_dem = np.count_nonzero(np.isnan(dem))
        if missing_dem > 0:
            log.warning(f"--> Filling {missing_dem} missing pixels in 'z'.")
            dem[np.isnan(dem)] = 0.0

        dem_coarse = local_mean(dem, pixel_size, radius)
        t_diff = (dem - dem_coarse) * lapse_rate
        ds["t_diff"] = xr.DataArray(t_diff, 
                                    coords = {
                                                "y": ds.y, 
                                                "x": ds.x})

        ds["t_diff"].attrs = {"sources": ["temperature.lapse_rate()"]}

        log.sub()

    new_data = ds[var] + ds["t_diff"]

    if "sources" in ds[var].attrs.keys():
        new_data.attrs = {"sources": ds[var].attrs["sources"]}

    if not isinstance(out_var, type(None)):
        ds[out_var] = new_data
    else:
        ds[var] = new_data

    return ds

def create_circle(pixel_size_x, pixel_size_y, radius, boolean = False):
    """Create a 2d array with pixel values representing the distance to 
    the center of the array.

    Parameters
    ----------
    pixel_size_x : float
        Size of a pixel in the x-direction. Unit can be anything, as long as it
        is the same as for `pixel_size_y` and `radius`.
    pixel_size_y : float
        Size of a pixel in the y-direction. Unit can be anything, as long as it
        is the same as for `pixel_size_x` and `radius`.
    radius : float
        Size of the circle. Unit can be anything, as long as it
        is the same as for `pixel_size_x` and `pixel_size_y`.
    boolean : bool, optional
        Return distances to center (False) or whether or not the distance to 
        the center is smaller than `radius` (True), by default False.

    Returns
    -------
    ndarray
        Values representing the distance to the center of the array.
    """
    min_pixel_size = np.min([np.abs(pixel_size_y), np.abs(pixel_size_x)])

    length = int(np.ceil(radius / min_pixel_size)) + 1

    # Y, X = np.ogrid[:length, :length]  * np.array([pixel_size_x, pixel_size_y])

    Y = np.rot90((np.arange(0, length, 1) * pixel_size_x)[np.newaxis,...], -1)
    X = (np.arange(0, length, 1) * pixel_size_y)[np.newaxis,...]

    center = (Y[-1, -1], X[-1,-1])

    dist = np.sqrt((X - center[1])**2 + (Y-center[0])**2)

    half_circle = np.concatenate([dist, np.flipud(dist)[1:,:]])
    circle = np.concatenate([half_circle, np.fliplr(half_circle)[:,1:]], axis =1)
    
    if boolean:
        circle = (circle <= radius).astype(int)
    
    return circle

def local_mean(array, pixel_size, radius, method = 3):
    """Calculate local means.

    Parameters
    ----------
    array : ndarray
        Two dimensional array for which local means will be calculated.
    pixel_size : tuple
        Size of the pixels in `array` in x and y direction respectively.
    radius : float
        Size of the circle around each pixel from which to 
        calculate the local mean.

    Returns
    -------
    ndarray :
        Array with the local mean of each pixel.
    """

    pixel_size_x, pixel_size_y = pixel_size
    circle_weights = create_circle(pixel_size_x, pixel_size_y, 
                                    radius, boolean=True)

    if method == 1: # *very* slow, i.e. takes more than 100 times longer than 2 and 3.
        divisor = convolve2d(np.ones_like(array), circle_weights, mode = "same")
        array_coarse = convolve2d(array, circle_weights, mode = "same") / divisor

    if method == 2:
        divisor = oaconvolve(np.ones_like(array), circle_weights, mode = "same")
        array_coarse = oaconvolve(array, circle_weights, mode = "same") / divisor

    if method == 3: # this one seems to be the fastest, but almost the same as method2.
        divisor = fftconvolve(np.ones_like(array), circle_weights, mode = "same")
        array_coarse = fftconvolve(array, circle_weights, mode = "same") / divisor


    return array_coarse

def bt_to_lst(ds, x, ndvi_s = 0.2, ndvi_v = 0.5, emis_s = 0.97, emis_v = 0.985):
    """Convert brightness temperature to land surface temperature according to Sobrino et al (2008).

    Parameters
    ----------
    ds : xr.Dataset
        Input.
    x : None
        Not used.
    ndvi_s : float, optional
        Threshold value, by default 0.2
    ndvi_v : float, optional
        Threshold value, by default 0.5
    emis_s : float, optional
        Threshold value, by default 0.97
    emis_v : float, optional
        Threshold value, by default 0.985

    Returns
    -------
    xr.Dataset
        Dataset in which `bt` has been replaced with `lst`.
    """

    if np.all([var in ds.data_vars for var in ["ndvi", "bt"]]):

        fvc = ((ds["ndvi"] - ndvi_s) / (ndvi_v - ndvi_s))**2
        emis = emis_s + (emis_v - emis_s) * fvc
        emis = xr.where(ds["ndvi"] < ndvi_s, emis_s, emis)
        emis = xr.where(ds["ndvi"] > ndvi_v, emis_v, emis)
        new_lst = ds["bt"] / emis

        if np.all([var in ds.data_vars for var in ["lst"]]):
            ds["lst"] = ds["lst"].fillna(new_lst)
        else:
            ds["lst"] = new_lst

        ds = ds.drop_vars(["bt"])

    return ds

# if __name__ == "__main__":

#     import numpy as np
#     import cProfile
#     import time

#     # fast case
#     # array = np.random.rand(800, 800)
#     # pixel_size = (0.01, 0.01)
#     # radius = 0.25    

#     # real case
#     array = np.random.rand(1009, 807)
#     pixel_size = (9.92063500e-04, -9.92063500e-04)
#     radius = 0.25

#     print("\n method2")
#     start = time.time()
#     result2 = local_mean(array, pixel_size, radius, method = 2)
#     end = time.time()
#     dt = end - start
#     print(dt)

#     print("\n method3")
#     start = time.time()
#     result3 = local_mean(array, pixel_size, radius, method = 3)
#     end = time.time()
#     dt = end - start
#     print(dt)

    # print("\n method1")
    # start = time.time()
    # result1 = local_mean(array, pixel_size, radius, method = 1)
    # end = time.time()
    # dt = end - start
    # print(dt)

    # import matplotlib.pyplot as plt

    # plt.figure(1)
    # plt.imshow(result1 - result2)

    # plt.figure(2)
    # plt.imshow(result1 - result3)






# -*- coding: utf-8 -*-

from pywapor.et_look_dev import constants as c

def lue(lue_max, stress_temp, stress_moist, eps_a):
    r"""
    Computes the light use efficiency

    Parameters
    ----------
    lue_max : float
        maximal light use efficiency
        :math:`EF_{ref}`
        [gr/MJ]
    stress_temp : float
        stress factor for air temperature
        :math:`S_{T}`
        [-]
    stress_moist : float
        stress soil moisture
        :math:`I`
        [-]
    eps_a : float
        epsilon autotrophic respiration
        :math:`I`
        [-]
        
    Returns
    -------
    lue : float
        light use efficiency
        :math:`I`
        [gr/MJ]
    """   
    
    lue = lue_max * stress_temp * stress_moist * eps_a

    return lue
   
def biomass(apar, lue):     
    r"""
    Computes the light use efficiency

    Parameters
    ----------
    apar : float
        apar
        :math:`I_{ra_24,fpar}`
        [MJ/m2]
    lue : float
        light use efficiency
        :math:`I`
        [gr/MJ]
        
    Returns
    -------
    biomass : float
        biomass production
        :math:`I`
        [kg/ha]
    """  
    
    biomass = apar * lue * 10
    
    return biomass


import numpy as np

def adjust_wind_height(ds, var, z = 10):
    """Adjust wind height from `z` to 2m according to FAO56 equation47.

    Parameters
    ----------
    ds : xr.Dataset
        Input
    var : str
        Variable to adjust.
    z : int, optional
        wind height in `ds`, by default 10.

    Returns
    -------
    xr.Dataset
        Dataset in which `var` has been adjusted.
    """

    if "2m" in var:
        old_name = var.replace("2m", f"{z}m")
    else:
        old_name = var
    ds[var] = ds[old_name] * ((4.87)/(np.log(67.8*z - 5.42)))
    return ds

def windspeed(ds, var):
    """Calculate windspeed from `u` and `v` wind vectors.

    Parameters
    ----------
    ds : xr.Dataset
        Input.
    var : str
        Variable in which to store the windspeed.

    Returns
    -------
    xr.Dataset
        Dataset with added `var`.
    """
    if np.all([x in ds.data_vars for x in ["u10m", "v10m"]]):
        ds[var] = np.hypot(ds["u10m"], ds["v10m"])
    return ds




import os
import pywapor
import cdsapi
import logging
import pandas as pd
import numpy as np
import glob
import xarray as xr
import rasterio.crs
import itertools
import copy
import re
from pywapor.enhancers.apply_enhancers import apply_enhancer
from pywapor.general.logger import log, adjust_logger
import shutil
from pywapor.general.processing_functions import save_ds

def create_time_settings(timelim):
    """Reformats the time limits so that they can be ingested by CDS.

    Parameters
    ----------
    timelim : list
        Period for which to prepare data.

    Returns
    -------
    list
        Time entries for CDS.
    """

    dates = pd.date_range(timelim[0], timelim[1], freq = "D")
    settings = list()
    for yr in np.unique(dates.year):
        for mnth in np.unique(dates.month[dates.year == yr]):
            days = dates.day[np.all([dates.year == yr, dates.month == mnth], axis = 0)]
            settings.append({"year": f"{yr}", "month": f"{mnth:02d}", "day": [f"{x:02d}" for x in days]})
    return settings

def request_size(setting):
    """Check the total size of a CDS request.

    Parameters
    ----------
    setting : list
        Time entries for CDS.

    Returns
    -------
    float
        Size of request.
    """
    relevant = ["day", "month", "year", "time", "variable"]
    size = np.prod([len(setting[selector]) for selector in relevant if isinstance(setting.get(selector), list)])
    return size

def split_setting(setting, max_size = 100):
    """Split a request into smaller requests untill size is smaller than `max_size`.

    Parameters
    ----------
    setting : list
        Time entries for CDS.
    max_size : int, optional
        Max allowed request size by CDS, by default 100.

    Returns
    -------
    list
        Setting broken up into smaller parts.
    """
    size = request_size(setting)
    if size <= max_size:
        new_settings = [setting]
    else:
        new_settings = list()
        for split_by in ['variable', 'time', 'year', 'month', 'day']:
            if isinstance(setting.get(split_by), list):
                for x in setting[split_by]:
                    new_setting = copy.copy(setting)
                    new_setting[split_by] = x
                    new_settings.append(new_setting)
                break
    return new_settings

def split_settings(settings, max_size = 100):
    """Split multuple requests into smaller requests untill all sizes are smaller than `max_size`.

    Parameters
    ----------
    setting : list
        Time entries for CDS.
    max_size : int, optional
        Max allowed request size by CDS, by default 100.

    Returns
    -------
    list
        Settings broken up into smaller parts.
    """
    while np.max([request_size(setting) for setting in settings]) > max_size:
        settings = list(itertools.chain.from_iterable([split_setting(setting, max_size = max_size) for setting in settings]))
    return settings

def download(folder, product_name, latlim, lonlim, timelim, variables, post_processors):
    """Download data from CDS.

    Parameters
    ----------
    folder : str
        Path to folder in which to save data.
    product_name : str
        Product name to download.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    variables : dict
        keys are variable names, values are additional settings.
    post_processors : dict
        processors to apply to specific variables.

    Returns
    -------
    xr.Dataset
        Dataset with downloaded data.
    """

    fn_final = os.path.join(folder, f"{product_name}.nc")

    # Create the settings for each individual request.
    time_settings = create_time_settings(timelim)
    area_settings = {"area": [latlim[1], lonlim[0], latlim[0], lonlim[1]]}
    settings = list()
    for var, extra_settings in variables.items():
        for t_setting in time_settings:
            settings.append({**t_setting, **extra_settings[0], 
                                **{"variable": var}, **area_settings})

    # Make sure the the individual request don't exceed the max allowed request size.
    max_size = {"sis-agrometeorological-indicators": 100}.get(product_name)
    if isinstance(max_size, int):
        settings = split_settings(settings, max_size = max_size)

    # Load api key.
    url, key = pywapor.collect.accounts.get("ECMWF")

    def info_callback(*args, **kwargs):
        _ = log.add().debug(*args, **kwargs).sub()

    def warning_callback(*args, **kwargs):
        _ = log.add().warning(*args, **kwargs).sub()

    def error_callback(*args, **kwargs):
        _ = log.add().error(*args, **kwargs).sub()

    def debug_callback(*args, **kwargs):
        _ = log.add().debug(*args, **kwargs).sub()

    _ = log.info("--> Directing CDS logging to file.")

    # Connect to server.
    c = cdsapi.Client(url = url, key = key, verify = True, 
                        info_callback=info_callback,
                        warning_callback=warning_callback,
                        error_callback=error_callback,
                        debug_callback=debug_callback,
                        quiet = True
                        )

    c.progress = True

    dss = list()
    subfolders = list()

    # Loop over requests
    for setting in settings:

        ext = {"zip": "zip", "netcdf": "nc", "grib": "grib", "tgz": "tar.gz"}[setting["format"]]
        fn = f"{setting['year']}_{setting['month']}_{setting['variable']}_{product_name}"
        fp = os.path.join(folder, f"{fn}.{ext}")

        # Make the request
        if not os.path.isfile(fp):
            _ = c.retrieve(product_name, setting, fp)

        # Unpack if necessary
        if ext in ["zip", "tar.gz"]:
            subfolder = os.path.join(folder, fn)
            if not os.path.exists(subfolder):
                os.makedirs(subfolder)
            shutil.unpack_archive(fp, subfolder)
            fps = glob.glob(os.path.join(subfolder, "*.nc"))
            subfolders.append(subfolder)
        else:
            fps = [fp]

        # Open downloaded data
        ds = xr.open_mfdataset(fps)

        das = list()

        time_offset = {"sis-agrometeorological-indicators": 12,
                "reanalysis-era5-single-levels": 0}[product_name]

        for var in ds.data_vars:
            # Fix time of relative humidity in agERA5.
            if bool(re.search(r'_[01]\dh', var)):
                offset = int(re.search(r'_[01]\dh', var).group()[1:-1])
                da = ds[var].assign_coords({"time": ds[var].time + np.timedelta64(offset, "h")})
            # Adjust time to middle of day for daily data.
            else:
                da = ds[var].assign_coords({"time": ds[var].time + np.timedelta64(time_offset, "h")})
            das.append(da)

        ds = xr.concat(das, dim="time").to_dataset().sortby("time")

        renames = {x: variables[setting["variable"]][1] for x in ds.data_vars}
        ds = ds.rename_vars(renames)

        dss.append(ds)

    # Merge everything together.
    ds = xr.merge(dss)

    # Clean up the dataset.
    relevant_coords = {
        "lat": "y", 
        "latitude": "y", 
        "lon": "x", 
        "longitude": "x", 
        # "time": "time",
    }

    coord_renames = {k: v for k, v in relevant_coords.items() if k in ds.coords}
    ds = ds.rename_dims(coord_renames)
    ds = ds.rename_vars(coord_renames)
    ds = ds.drop_vars([x for x in ds.coords if x not in ds.dims])
    ds = ds.rio.write_crs(rasterio.crs.CRS.from_epsg(4326))
    ds = ds.rio.write_grid_mapping("spatial_ref")
    for var in list(ds.data_vars):
        ds[var].attrs = {k:v for k,v in ds[var].attrs.items() if k == "units"}
    ds = ds.sortby("y", ascending = False)
    ds = ds.sortby("x")
    ds.attrs = {}

    # Apply product specific functions.
    for var, funcs in post_processors.items():
        for func in funcs:
            ds, label = apply_enhancer(ds, var, func)
            log.info(label)

    # Save the netcdf.
    ds = save_ds(ds, fn_final, label = "Merging files.")

    # Remove unpacked zips.
    for subfolder in subfolders:
        if os.path.isdir(subfolder):
            shutil.rmtree(subfolder)
    
    return ds

# if __name__ == "__main__":

#     folder = r"/Users/hmcoerver/On My Mac/era_test"
#     latlim = [28.9, 29.7]
#     lonlim = [30.2, 31.2]
#     timelim = ["2022-04-01", "2022-04-10"]

#     adjust_logger(True, folder, "INFO")

#     product_name = "sis-agrometeorological-indicators"
#     # product_name = "reanalysis-era5-single-levels"

#     req_vars = ["t_air", "t_dew", ]#"rh", "u", "vp", "ra"]
#     # req_vars = ["u_10m", "v_10m", "t_dew", "p_air_0", "p_air", "t_air"]

#     variables = pywapor.collect.product.ERA5.default_vars(product_name, req_vars)

#     download(folder, product_name, latlim, lonlim, timelim, variables)

#     _ = log.info("test")